--[[
Orginal Script: HighPressureWasherPlaceable.lua by GIANTS Software GmbH

Rewritting for buyable vehilces: Ifko[nator]
Date: 14.02.2016
Version: 1.0

You must this HPW fillup with water to use!

Name: pressureWasher.lua
]]

pressureWasher = {};

function pressureWasher.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization((Fillable and WaterTrailer), specializations);
end;

function pressureWasher:load(xmlFile)
	self.isInActionDistance = pressureWasher.isInActionDistance;
	self.getIsPlayerInRange = pressureWasher.getIsPlayerInRange;
	self.setIsWashing = pressureWasher.setIsWashing;
	self.setIsTurnedOn = pressureWasher.setIsTurnedOn;
	self.onDeactivate = pressureWasher.onDeactivate;
	self.activateLance = pressureWasher.activateLance;
	self.deactivateLance = pressureWasher.deactivateLance;
	self.deleteLance = pressureWasher.deleteLance;
	self.drawLance = pressureWasher.drawLance;
	self.updateLance = pressureWasher.updateLance;
	self.getPressureWasherActiveForSound = pressureWasher.getPressureWasherActiveForSound;
	self.cleanVehicle = pressureWasher.cleanVehicle;
	self.updateTickLance = pressureWasher.updateTickLance;
	self.washRaycastCallback = pressureWasher.washRaycastCallback;

	self.playerInRangeDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.playerInRangeDistance"), 3);
	self.washDistance = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.washDistance"), 10);
	self.actionRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.actionRadius#distance"), 15);
	self.washMultiplier = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.washMultiplier"), 1);
	self.waterUsePerSecond = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.waterUsePerMinute"), 2);
	self.lanceNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.lance#index"));
	self.linkPosition = Utils.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, "vehicle.lance#position"), "0 0 0"), 3);
	self.linkRotation = Utils.getRadiansFromString(Utils.getNoNil(getXMLString(xmlFile, "vehicle.lance#rotation"), "0 0 0"), 3);
	self.lanceNodeParent = getParent(self.lanceNode);
	self.lanceRaycastNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.lance#raycastNode"));
	
	self.messageShown = false;

	if self.isClient then
		self.particleSystems = {};

		Utils.loadParticleSystem(xmlFile, self.particleSystems, "vehicle.particleSystem", self.components, false, nil, g_currentMission.baseDirectory, nil);

		self.waterEffect = EffectManager:loadEffect(xmlFile, "vehicle.waterEffect", self.components, self);
		self.sampleCompressor = Utils.loadSample(xmlFile, {}, "vehicle.compressorSound", nil, self.baseDirectory, self.rootNode);
		self.compressorPitchMin = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.compressorSound#pitchMin"), 0.5);
		self.sampleWashing = Utils.loadSample(xmlFile, {}, "vehicle.washingSound", nil, self.rootNode);
		self.sampleSwitch = Utils.loadSample(xmlFile, {}, "vehicle.switchSound", nil, self.rootNode);
		local filename = getXMLString(xmlFile, "vehicle.exhaust#filename");

		if filename ~= nil then
			local i3dNode = Utils.loadSharedI3DFile(filename, self.baseDirectory, false, false, false);

			if i3dNode ~= 0 then
				local linkNode = Utils.getNoNil(Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.exhaust#index")), self.components);
				self.exhaustFilename = filename;
				self.exhaustNode = getChildAt(i3dNode, 0);

				link(linkNode, self.exhaustNode);
				setVisibility(self.exhaustNode, false);
				delete(i3dNode);
			end;
		end;

		self.targets = {};

		HandTool.loadTargets(xmlFile, "vehicle", self.components, self.targets);
	end;

	self.isPlayerInRange = false;
	self.isTurnedOn = false;
	self.doWashing = false;
	self.lastInRangePosition = {
		0,
		0,
		0
	};
	self.isTurningOff = false;
	self.turnOffTime = 0;
	self.turnOffDuration = 500;
end;

function pressureWasher:delete()
	self:setIsTurnedOn(false, nil, false);

	if self.isClient then
		EffectManager:deleteEffect(self.waterEffect);

		if self.exhaustFilename ~= nil then
			Utils.releaseSharedI3DFile(self.exhaustFilename, self.baseDirectory, true);
		end;

		if self.particleSystems ~= nil then
			Utils.deleteParticleSystem(self.particleSystems);
		end;

		if self.sampleCompressor ~= nil then
			Utils.deleteSample(self.sampleCompressor);
		end;

		if self.sampleWashing ~= nil then
			Utils.deleteSample(self.sampleWashing);
		end;

		if self.sampleSwitch ~= nil then
			Utils.deleteSample(self.sampleSwitch);
		end;
	end;

	unregisterObjectClassName(self);
end;

function pressureWasher:activateHandtool(player)
	self:setIsTurnedOn(true, player, true);
end;

function pressureWasher:update(dt)
	if self.currentPlayer ~= nil then
		local isPlayerInRange = self:getIsPlayerInRange(self.actionRadius, self.currentPlayer);
		local messageInRange = self:getIsPlayerInRange(self.actionRadius + 1, self.currentPlayer);

		if messageInRange and self.currentPlayer == g_currentMission.player then
			g_currentMission:addExtraPrintText(string.gsub(g_i18n:getText("CurrentFillLevel"), "currentFillLevel", math.floor(self.fillLevel)));
		end
		
		if isPlayerInRange then
			self.lastInRangePosition = {
				getTranslation(self.currentPlayer.rootNode)
			};
		else
			local kx, _, kz = getWorldTranslation(self.rootNode);
			local px, _, pz = getWorldTranslation(self.currentPlayer.rootNode);
			local len = Utils.vector2Length(px - kx, pz - kz);
			local x, y, z = unpack(self.lastInRangePosition);
			x = kx + (px - kx)/len*(self.actionRadius - dt * 1e-05);
			z = kz + (pz - kz)/len*(self.actionRadius - dt * 1e-05);
			y = math.max(y, getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x, 0, z));

			self.currentPlayer:moveToAbsoluteInternal(x, y, z);

			self.lastInRangePosition = {
				x,
				y,
				z
			};

			if not self.messageShown and self.currentPlayer == g_currentMission.player then
				g_currentMission:showBlinkingWarning(g_i18n:getText("HighPressureWasherRangeRestriction"), 6000);

				self.messageShown = true;
			end;
		end;
	end;
	
	if self.isTurnedOn and self.doWashing then
		if self.fillLevel > 0 then	
			self.foundVehicle = nil;
	
			self:cleanVehicle(self.currentPlayer.cameraId, dt);
	
			if self.lanceRaycastNode ~= nil then
				self:cleanVehicle(self.lanceRaycastNode, dt);
			end;

			self:setFillLevel(self.fillLevel - self.waterUsePerSecond, Fillable.FILLTYPE_WATER);
		else
			if self.currentPlayer == g_currentMission.player then	
				g_currentMission:showBlinkingWarning(g_i18n:getText("HighPressureWasherIsEmpty"), 4000);
			end;
			
			self.doWashing = false;
		end;
	end;

	if self.isTurningOff then
		if g_currentMission.time < self.turnOffTime then
			if self.sampleCompressor ~= nil and self:getIsPlayerInRange() then
				local pitch = Utils.lerp(self.compressorPitchMin, self.sampleCompressor.pitchOffset, Utils.clamp((self.turnOffTime - g_currentMission.time) / self.turnOffDuration, 0, 1));
				local volume = Utils.lerp(0, self.sampleCompressor.volume, Utils.clamp((self.turnOffTime - g_currentMission.time) / self.turnOffDuration, 0, 1));

				Utils.setSamplePitch(self.sampleCompressor, pitch);
				Utils.setSampleVolume(self.sampleCompressor, volume);
			end;
		else
			self.isTurningOff = false;

			if self.sampleCompressor ~= nil then
				Utils.stop3DSample(self.sampleCompressor);
			end;
		end;
	end;

	local isPlayerInRange, player = self:getIsPlayerInRange(self.playerInRangeDistance);

	if isPlayerInRange and player == g_currentMission.player then
		self.playerInRange = player;
		self.isPlayerInRange = true;
		local activateText = "unknown";

		if self.isTurnedOn then
			activateText = string.format(g_i18n:getText("turn_off_OBJECT"), g_i18n:getText("TypeDesc_HighPressureWasher"));
			
			if InputBinding.hasEvent(InputBinding.ACTIVATE_OBJECT) then
				self:setIsTurnedOn(false, nil);
			end;
		else
			activateText = string.format(g_i18n:getText("turn_on_OBJECT"), g_i18n:getText("TypeDesc_HighPressureWasher"));

			if InputBinding.hasEvent(InputBinding.ACTIVATE_OBJECT) then
				self:setIsTurnedOn(true, g_currentMission.player);
			end;
		end;
		
		g_currentMission:addHelpButtonText(activateText, InputBinding.ACTIVATE_OBJECT);
	else
		self.playerInRange = nil;
		self.isPlayerInRange = false;
	end;
end;

function pressureWasher:getIsPlayerInRange(distance, player)
	if self.rootNode ~= 0 then
		distance = Utils.getNoNil(distance, 10);

		if player == nil then
			for _, player in pairs(g_currentMission.players) do
				if self:isInActionDistance(player, self.rootNode, distance) then
					return true, player;
				end;
			end;
		else
			return self:isInActionDistance(player, self.rootNode, distance), player;
		end;
	end;

	return false, nil;
end;

function pressureWasher:isInActionDistance(player, refNode, distance)
	local x, y, z = getWorldTranslation(refNode);
	local px, py, pz = getWorldTranslation(player.rootNode);
	local dx = px - x;
	local dz = pz - z;

	if dx * dx + dz * dz < distance * distance then
		return true;
	end;

	return false;
end;

function pressureWasher:readStream(streamId, connection)
	if connection.getIsServer(connection) then
		local isTurnedOn = streamReadBool(streamId);

		if isTurnedOn then
			local player = networkGetObject(streamReadInt32(streamId));

			if player ~= nil then
				self:setIsTurnedOn(isTurnedOn, player, true);
			end;
		end;
	end;
end;

function pressureWasher:writeStream(streamId, connection)
	if not connection.getIsServer(connection) then
		streamWriteBool(streamId, self.isTurnedOn);

		if self.isTurnedOn then
			streamWriteInt32(streamId, networkGetObjectId(self.currentPlayer));
		end;
	end;
end;

function pressureWasher:cleanVehicle(node, dt)
	local x, y, z = getWorldTranslation(node);
	local dx, dy, dz = localDirectionToWorld(node, 0, 0, -1);
	local lastFoundVehicle = self.foundVehicle;

	raycastAll(x, y, z, dx, dy, dz, "washRaycastCallback", self.washDistance, self, 12770);

	if self.foundVehicle ~= nil and lastFoundVehicle ~= self.foundVehicle then
		self.foundVehicle:setDirtAmount(self.foundVehicle:getDirtAmount() - (self.washMultiplier * dt) / self.foundVehicle.washDuration);
	end;
end;

function pressureWasher:setIsWashing(doWashing, force, noEventSend)
	HPWPlaceableStateEvent.sendEvent(self, doWashing, noEventSend);
	
	if self.doWashing ~= doWashing then
		if self.isClient then
			Utils.setEmittingState(self.particleSystems, doWashing and self.fillLevel > 0);

			if doWashing and self.fillLevel > 0 then
				EffectManager:startEffect(self.waterEffect);

				if self.sampleWashing ~= nil and self:getIsPlayerInRange() then
					Utils.playSample(self.sampleWashing, 0, 0, 1);
				end
			else
				if force and self.fillLevel > 0 then
					EffectManager:resetEffect(self.waterEffect);
				else
					EffectManager:stopEffect(self.waterEffect);
				end;

				if self.sampleWashing ~= nil then
					Utils.stopSample(self.sampleWashing, true);
				end;
			end;
		end;

		self.doWashing = doWashing;
	end;
end;

function pressureWasher:setIsTurnedOn(isTurnedOn, player, noEventSend)
	HPWPlaceableTurnOnEvent.sendEvent(self, isTurnedOn, player, noEventSend);
	
	if self.isTurnedOn ~= isTurnedOn then
		if isTurnedOn then
			self.isTurnedOn = isTurnedOn;
			self.currentPlayer = player;
			local tool = {
				node = self.lanceNode
			};

			link(player.toolsRootNode, tool.node);
			setVisibility(tool.node, false);
			setTranslation(tool.node, unpack(self.linkPosition));
			setRotation(tool.node, unpack(self.linkRotation));

			tool.update = self.updateLance;
			tool.updateTick = self.updateTickLance;
			tool.delete = self.deleteLance;
			tool.draw = self.drawLance;
			tool.onActivate = self.activateLance;
			tool.onDeactivate = self.deactivateLance;
			tool.targets = self.targets;
			tool.owner = self;
			tool.static = false;
			self.tool = tool;

			self.currentPlayer:setTool(tool);

			self.currentPlayer.hasHPWLance = true;

			if self.isClient then
				if self.sampleSwitch ~= nil and self:getPressureWasherActiveForSound() then
					Utils.playSample(self.sampleSwitch, 1, 0, nil);
				end;

				if self.sampleCompressor ~= nil then
					Utils.setSamplePitch(self.sampleCompressor, self.sampleCompressor.pitchOffset);
					Utils.setSampleVolume(self.sampleCompressor, self.sampleCompressor.volume);
					Utils.play3DSample(self.sampleCompressor);
				end;

				if self.isTurningOff then
					self.isTurningOff = false;
				end;

				--setVisibility(self.lanceNode, g_currentMission.player == player);
				setVisibility(self.lanceNode, true);
			end;
		else
			self:onDeactivate();
		end;

		if self.exhaustNode ~= nil then
			setVisibility(self.exhaustNode, isTurnedOn);
		end;
	end;
end;

function pressureWasher:onDeactivate()
	if self.isClient and self.sampleSwitch ~= nil and self:getPressureWasherActiveForSound() then
		Utils.playSample(self.sampleSwitch, 1, 0, nil);
	end;

	self.isTurnedOn = false;

	setVisibility(self.lanceNode, true);
	self.setIsWashing(self, false, true, true);

	if self.currentPlayer ~= nil then
		self.currentPlayer:setToolById(0, true);

		self.currentPlayer.hasHPWLance = false;
	end;

	if self.isClient then
		if self.sampleWashing ~= nil then
			Utils.stopSample(self.sampleWashing, true);
		end;

		self.isTurningOff = true;
		self.turnOffTime = g_currentMission.time + self.turnOffDuration;

		link(self.lanceNodeParent, self.lanceNode);
		setTranslation(self.lanceNode, 0, 0, 0);
		setRotation(self.lanceNode, 0, 0, 0);
	end;

	self.currentPlayer = nil; 
end;

function pressureWasher:getPressureWasherActiveForSound()
	return self.isTurnedOn and self.currentPlayer == g_currentMission.player and g_gui.currentGui == nil;
end;

function pressureWasher:washRaycastCallback(hitObjectId, x, y, z, distance)
	local vehicle = g_currentMission.nodeToVehicle[hitObjectId];

	if vehicle ~= nil and vehicle.getDirtAmount ~= nil and vehicle.setDirtAmount ~= nil and vehicle.washDuration ~= nil then
		self.foundCoords = {
			x,
			y,
			z
		};
		self.foundVehicle = vehicle;

		return false;
	end;

	return true;
end;

function pressureWasher.activateLance(tool)
	setVisibility(tool.node, true); 
end;

function pressureWasher.deactivateLance(tool)
	tool.owner:setIsTurnedOn(false, nil);
end;

function pressureWasher.deleteLance(tool)
	tool.owner:setIsTurnedOn(false, nil);
end;

function pressureWasher.drawLance(tool)
	if tool.owner.currentPlayer == g_currentMission.player then
		g_currentMission:addHelpButtonText(g_i18n:getText("ACTIVATE_HANDTOOL"), InputBinding.ACTIVATE_HANDTOOL);
	end;
end;

function pressureWasher.updateLance(tool, dt, allowInput)
	if allowInput then
		tool.owner:setIsWashing(InputBinding.isPressed(InputBinding.ACTIVATE_HANDTOOL), false, false);
	end;
end;

--## unused functions

function pressureWasher:mouseEvent(posX, posY, isDown, isUp, button)end;
function pressureWasher:keyEvent(unicode, sym, modifier, isDown)end;
function pressureWasher:updateTick(dt)end;
function pressureWasher:draw()end;
function pressureWasher.updateTickLance(tool, dt, allowInput)end;